/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.hoau;

import java.applet.Applet;
import java.awt.Dimension;
import java.awt.Font;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.text.DecimalFormat;
import java.text.NumberFormat;

import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.text.BadLocationException;

import at.jku.risc.stout.hoau.algo.AntiUnify;
import at.jku.risc.stout.hoau.algo.AntiUnifyProblem;
import at.jku.risc.stout.hoau.algo.AntiUnifySystem;
import at.jku.risc.stout.hoau.algo.DebugLevel;
import at.jku.risc.stout.hoau.data.EquationSystem;
import at.jku.risc.stout.hoau.data.InputParser;
import at.jku.risc.stout.hoau.data.NodeFactory;
import at.jku.risc.stout.hoau.data.TermNode;
import at.jku.risc.stout.hoau.util.ControlledException;

/**
 * Simple presentation layer for embedding the rule based algorithm
 * {@linkplain AntiUnifySystem} in the browser.
 * 
 * @author Alexander Baumgartner
 */
public class AntiUnifyApplet extends Applet {
	private static final long serialVersionUID = 1L;

	private JTextArea textArea = new JTextArea();
	private JScrollPane scroll = new JScrollPane(textArea,
			JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
			JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);

	private static final int MAX_LINES = 5000;
	// Jitter for MAX_LINES = 4900 - 5100
	private static final int MAX_LINES_VARIANZ = 100;
	private static final NumberFormat LINENUM_FORMAT = new DecimalFormat(
			"000000: ");
	private static final String PRINT_HEAD = "";
	private static final String PRINT_TAIL = "DONE";

	private final PrintStream out = new PrintStream(new OutputStream() {
		StringBuilder buffer = new StringBuilder();
		long count = 0L;

		@Override
		public void write(int b) throws IOException {
			buffer.appendCodePoint(b);
			if (b == '\n') {
				count++;
				flush();
			}
		}

		@Override
		public void flush() throws IOException {
			display(LINENUM_FORMAT.format(count) + buffer.toString());
			buffer.setLength(0);
		}
	});
	private final PrintStream info = new PrintStream(new OutputStream() {
		StringBuilder buffer = new StringBuilder();

		@Override
		public void write(int b) throws IOException {
			buffer.appendCodePoint(b);
			if (b == '\n')
				flush();
		}

		@Override
		public void flush() throws IOException {
			display("INFO:  " + buffer.toString());
			buffer.setLength(0);
		}
	});
	private final PrintStream err = new PrintStream(new OutputStream() {
		StringBuilder buffer = new StringBuilder();

		@Override
		public void write(int b) throws IOException {
			buffer.appendCodePoint(b);
			if (b == '\n')
				flush();
		}

		@Override
		public void flush() throws IOException {
			display("ERROR: " + buffer.toString());
			buffer.setLength(0);
		}
	});

	@Override
	public void init() {
		textArea.setEditable(false);
		scroll.setPreferredSize(new Dimension(getWidth(), getHeight() - 5));
		textArea.setFont(new Font(Font.MONOSPACED, Font.PLAIN, getHeight() / 20));
		add(scroll);
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
			SwingUtilities.updateComponentTreeUI(this);
		} catch (Exception isOk_UseDefaultLookAndFeel) {
		}
	}

	@Override
	public void resize(int width, int height) {
		super.resize(width, height);
		scroll.setPreferredSize(new Dimension(width, height - 5));
		textArea.setFont(new Font(Font.MONOSPACED, Font.PLAIN, getHeight() / 20));
	}

	@Override
	public void start() {
		display(PRINT_HEAD);
		try {
			String unifProblem = getParameter("problemset");
			if (unifProblem == null || unifProblem.length() == 0) {
				err.println("No anti-unification problem sent by the browser.\n\n"
						+ "This might be a communication-error between your browser and the Java-Applet.\n"
						+ "Please try it once more. If you use Opera, please turn off the Turbo mode.\n"
						+ "If the problem persists, please try another browser.");
			} else {
				boolean justify = "true".equals(getParameter("justify"));
				boolean userFriendly = "true"
						.equals(getParameter("userfriendly"));
				DebugLevel debugLevel = DebugLevel.SIMPLE;
				int maxReduce = 100;
				try {
					debugLevel = DebugLevel.valueOf(getParameter("debugmode"));
					maxReduce = Integer.parseInt(getParameter("maxreduce"));
				} catch (Exception e) {
				}

				EquationSystem<AntiUnifyProblem> eqSys = new EquationSystem<AntiUnifyProblem>() {
					@Override
					public AntiUnifyProblem newEquation() {
						return new AntiUnifyProblem();
					}
				};
				TermNode.printUserFriendly = userFriendly;
				NodeFactory.resetCounter(); // reset fresh variable counter
				InputParser parser = new InputParser();
				parser.tryFreeVarPrefix(unifProblem, "X", "Y", "Z", "V", "U");
				parser.tryBoundVarPrefix(unifProblem, "x", "y", "z", "v", "u");

				info.println("Justify computed generalization = " + justify);
				info.println("Maximum reduction recursion = " + maxReduce);
				info.println("Output format = " + debugLevel
						+ " (User friendly = " + userFriendly + ")");
				info.println();
				info.println("PROBLEM:");
				parser.parseEquationSystem(unifProblem, eqSys, maxReduce);
				info.println(eqSys);
				info.println();

				AntiUnify au = new AntiUnify(eqSys, maxReduce, debugLevel);
				au.antiUnify(justify, out);
			}
		} catch (ControlledException ex) {
			err.println(ex.getMessage() + " (" + ex.getClass().getSimpleName()
					+ ")");
		} catch (Throwable ex) {
			ex.printStackTrace(err);
		}
		display(PRINT_TAIL);
		try {
			Thread.sleep(300);
		} catch (InterruptedException ignored) {
		}
	}

	private void display(String str) {
		try {
			if (textArea.getLineCount() >= MAX_LINES + MAX_LINES_VARIANZ) {
				textArea.replaceRange(null, 0,
						textArea.getLineEndOffset(MAX_LINES_VARIANZ * 2));
			}
			textArea.append(str);
		} catch (BadLocationException e) {
			e.printStackTrace();
		}
	}

	/**
	 * May be called by JavaScript to enable copying of text from the applet to
	 * the system clipboard.
	 */
	public String getSelectedText() {
		String text = textArea.getSelectedText();
		return text == null ? "" : text;
	}
}
