/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.hoau;

import java.io.PrintStream;
import java.util.Arrays;

import at.jku.risc.stout.hoau.algo.AntiUnify;
import at.jku.risc.stout.hoau.algo.AntiUnifyProblem;
import at.jku.risc.stout.hoau.algo.AntiUnifySystem;
import at.jku.risc.stout.hoau.algo.DebugLevel;
import at.jku.risc.stout.hoau.data.EquationSystem;
import at.jku.risc.stout.hoau.data.InputParser;
import at.jku.risc.stout.hoau.data.NodeFactory;
import at.jku.risc.stout.hoau.data.TermNode;

/**
 * Simple presentation layer for running the rule based algorithm
 * {@linkplain AntiUnifySystem} in a shell.
 * 
 * @author Alexander Baumgartner
 */
public class AntiUnifyMain {

	static DebugLevel debugLevel = DebugLevel.SIMPLE;
	static boolean userFriendly = false;
	static boolean justifySigma = false;
	static int maxRec = 100;

	public static void main(String[] args) {
		if (args.length >= 1) {
			if ("-o".equals(args[0])) {
				debugLevel = DebugLevel.PROGRESSORIGIN;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-d".equals(args[0])) {
				if (debugLevel != DebugLevel.PROGRESSORIGIN)
					debugLevel = DebugLevel.PROGRESS;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-v".equals(args[0])) {
				if (debugLevel != DebugLevel.PROGRESS
						&& debugLevel != DebugLevel.PROGRESSORIGIN)
					debugLevel = DebugLevel.VERBOSE;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-u".equals(args[0])) {
				userFriendly = true;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-h".equals(args[0])) {
				printHelp(System.out);
			} else if ("-j".equals(args[0])) {
				justifySigma = true;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-m".equals(args[0])) {
				if (args.length < 2) {
					System.out.println("Insufficient parameters!");
					System.out.println();
					printHelp(System.out);
				} else {
					maxRec = Integer.parseInt(args[1]);
					main(Arrays.copyOfRange(args, 2, args.length));
				}
			} else {
				try {
					PrintStream out = System.out;
					EquationSystem<AntiUnifyProblem> eqSys = new EquationSystem<AntiUnifyProblem>() {
						@Override
						public AntiUnifyProblem newEquation() {
							return new AntiUnifyProblem();
						}
					};
					TermNode.printUserFriendly = userFriendly;
					NodeFactory.resetCounter();
					InputParser parser = new InputParser();
					parser.tryFreeVarPrefix(args[0], "X", "Y", "Z", "V", "U");
					parser.tryBoundVarPrefix(args[0], "x", "y", "z", "v", "u");

					out.println("Justify computed generalization = "
							+ justifySigma);
					out.println("Maximum reduction recursion = " + maxRec);
					out.println("Output format = " + debugLevel
							+ " (User friendly = " + userFriendly + ")");
					out.println();
					out.println("PROBLEM:");
					parser.parseEquationSystem(args[0], eqSys, maxRec);
					out.println();

					AntiUnify au = new AntiUnify(eqSys, maxRec, debugLevel);
					au.antiUnify(justifySigma, out);

					System.out.println();
					System.out
							.println("Supported by the Austrian Science Fund (FWF) "
									+ "under the project SToUT P 24087-N18");
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		} else {
			System.out.println("Insufficient parameters!");
			System.out.println();
			printHelp(System.out);
		}
	}

	private static void printHelp(PrintStream out) {
		System.out.println("Anti-unification problem solver for higher order "
				+ "pattern");
		System.out.println();
		System.out.println("Give a problem of the form:");
		System.out.println("\"s1 =^= t1; s2 =^= t2; ...; sn =^= tn\"");
		System.out
				.println("where s1,...,sn and t1,...,tn are higher order pattern");
		System.out.println("Input example: \"\\x,y.f(x,y) =^= \\x,y.f(y,x)\"");
		System.out
				.println("     Or typed: \"\\x:t1,y:t2.f(x,y) =^= \\x:t1,y:t2.f(y,x)\"");
		System.out.println();
		System.out.println("Simple execution example: "
				+ "java at.jku.risc.stout.hoau.AntiUnifyMain -u -j "
				+ "\"\\x.x =^= \\y.y\"");
		System.out.println("Advanced execution example with jvm parameters:");
		System.out.println("java -Xmx2048m -XX:MaxPermSize=1024m -Xss128m "
				+ "at.jku.risc.stout.hoau.AntiUnifyMain -j -m 1000000 "
				+ "\"\\x.x =^= \\y.y\"");
		System.out.println();
		System.out.println("  -j    justify computed generalization");
		System.out.println("  -u    user friendly form");
		System.out.println("  -m n  set maximum reduction recursion to n");
		System.out
				.println("  -v    verbose output mode prints sigma and the store");
		System.out.println("  -h    print this information");
		System.out
				.println("  -d    debug mode prints out every available information");
		System.out
				.println("  -o    debug mode prints out every available information with origin variable names");
	}
}
