/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.hoau.algo;

import java.util.List;
import java.util.Map;
import java.util.Set;

import at.jku.risc.stout.hoau.data.Equation;
import at.jku.risc.stout.hoau.data.Hedge;
import at.jku.risc.stout.hoau.data.LambdaHedge;
import at.jku.risc.stout.hoau.data.NodeFactory;
import at.jku.risc.stout.hoau.data.TermNode;
import at.jku.risc.stout.hoau.data.atom.BoundVariable;
import at.jku.risc.stout.hoau.data.atom.Lambda;
import at.jku.risc.stout.hoau.data.atom.TermAtom;
import at.jku.risc.stout.hoau.data.atom.Variable;
import at.jku.risc.stout.hoau.util.DataStructureFactory;

/**
 * This class represents an anti-unification problem (AUP) which consists of one
 * generalization variable (the most general generalization), a list of
 * abstracted bound variables and an anti-unification equation. The list of
 * abstracted variables is additionally stored inside a set for faster lookup.
 * 
 * @author Alexander Baumgartner
 */
public class AntiUnifyProblem implements Equation {
	private Variable generalizationVar = NodeFactory.obtainFreshVar(null);
	private List<BoundVariable> abstractionList = DataStructureFactory.$
			.newList();
	// HashSet has O(1) access
	private Set<BoundVariable> abstractionSet = DataStructureFactory.$.newSet();
	private TermNode left, right;
	public static String PRINT_EQ_SEPARATOR = " =^= ";
	public static String PRINT_VAR_SEPARATOR = ": ";

	@Override
	public AntiUnifyProblem clone() {
		try {
			AntiUnifyProblem clone = (AntiUnifyProblem) super.clone();
			clone.abstractionList = DataStructureFactory.$
					.newList(abstractionList);
			clone.abstractionSet = DataStructureFactory.$
					.newSet(abstractionSet);
			clone.left = left.clone();
			clone.right = right.clone();
			return clone;
		} catch (CloneNotSupportedException e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public String toString() {
		StringBuilder ret = new StringBuilder(512);
		ret.append(String.valueOf(generalizationVar));
		if (abstractionList.size() > 0)
			ret.append(String.valueOf(abstractionList));
		ret.append(PRINT_VAR_SEPARATOR);
		ret.append(String.valueOf(left));
		String rString = String.valueOf(right);
		if (rString.length() > 0)
			ret.append(PRINT_EQ_SEPARATOR).append(right);
		return ret.toString();
	}

	@Override
	public void setLeft(TermNode left) {
		this.left = left.reduceHedges();
	}

	@Override
	public void setRight(TermNode right) {
		this.right = right.reduceHedges();
	}

	@Override
	public TermNode getLeft() {
		return left;
	}

	@Override
	public TermNode getRight() {
		return right;
	}

	/**
	 * Returns the generalization variable, which represents the most general
	 * generalization of this AUP.
	 */
	public Variable getGeneralizationVar() {
		return generalizationVar;
	}

	/**
	 * Tests whether the set of abstracted variables contains the given
	 * variable.
	 */
	public boolean abstrContains(TermAtom boundVar) {
		return abstractionSet.contains(boundVar);
	}

	/**
	 * Returns the set of abstracted variables.
	 */
	public Set<BoundVariable> abstrSet() {
		return abstractionSet;
	}

	/**
	 * Returns the length of the list of abstractions.
	 */
	public int abstrSize() {
		return abstractionList.size();
	}

	/**
	 * Returns the abstracted bound variable corresponding to the given position
	 * index from the list of abstractions.
	 */
	public BoundVariable abstrItem(int index) {
		return abstractionList.get(index);
	}

	/**
	 * Adds a bound variable to the list of abstractions
	 */
	public void addAbstrVar(BoundVariable boundVar) {
		abstractionList.add(boundVar);
		abstractionSet.add(boundVar);
	}

	/**
	 * Copies the variables from the abstraction list of the given AUP into the
	 * abstraction list of this AUP.
	 */
	public void copyAbstr(AntiUnifyProblem eq) {
		abstractionList.addAll(eq.abstractionList);
		abstractionSet.addAll(eq.abstractionList);
	}

	/**
	 * Binds all the abstracted variables to the given term.
	 */
	public TermNode bindAll(TermNode t) {
		for (int i = abstractionList.size() - 1; i >= 0; i--) {
			LambdaHedge lh = new LambdaHedge();
			lh.add(t);
			t = new TermNode(new Lambda(abstractionList.get(i)), lh);
		}
		return t;
	}

	/**
	 * X(x,y,...) where X is the generalization variable and x,y,... are the
	 * variables in the abstraction list.
	 */
	public TermNode createHO_Fnc() {
		return createHO_Fnc(null);
	}

	/**
	 * X(x,y,...) where X is the generalization variable and x,y,... are the
	 * variables in the abstraction list and pi is a variable renaming for
	 * x,y,...
	 */
	public TermNode createHO_Fnc(Map<Variable, Variable> pi) {
		if (abstractionList.size() == 0)
			return new TermNode(generalizationVar, null);
		Hedge hedge = new Hedge();
		for (int i = 0, n = abstractionList.size(); i < n; i++) {
			TermAtom to = abstractionList.get(i);
			if (pi != null) {
				TermAtom tmp = pi.get(abstractionList.get(i));
				if (tmp != null)
					to = tmp;
			}
			hedge.add(new TermNode(to, null));
		}
		return new TermNode(generalizationVar, hedge);
	}

	/**
	 * Creates a lambda hedge which contains the given term node and all the
	 * abstracted bound variables. For instance if the list of abstracted
	 * variables contains x and y and the given term node consists of only one
	 * free variable X, then the created lambda hedge would be (X x y). When
	 * substituting X later with a lambda term like \z,v. f[a, z, v] the
	 * resulting term will become ((\z,v. f[a, z, v]) x y) and after beta
	 * reduction it will be f[a, x, y].
	 */
	public TermNode createLambdaHedge(TermNode lambda) {
		if (abstractionList.size() == 0)
			return lambda;
		Hedge lh = new LambdaHedge();
		lh.add(lambda);
		for (int i = 0, n = abstractionList.size(); i < n; i++) {
			lh.add(new TermNode(abstractionList.get(i), null));
		}
		return new TermNode(null, lh);
	}

	/**
	 * Abstracts the lambda bound variables of both sides of the equation of
	 * this AUP. The top atom of both sides has to be a lambda atom!
	 */
	public void abstrLambda() {
		BoundVariable lBound = ((Lambda) left.getAtom()).getBoundVar();
		BoundVariable rBound = ((Lambda) right.getAtom()).getBoundVar();
		left.setAtom(null);
		right.setAtom(null);
		left = left.reduceHedges();
		right = right.reduceHedges();
		right.substitute(rBound, lBound);
		addAbstrVar(lBound);
	}
}
