/*
 * Copyright 2012 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.hoau.algo;

import java.io.PrintStream;
import java.util.Map;
import java.util.Set;

import at.jku.risc.stout.hoau.data.EquationSystem;
import at.jku.risc.stout.hoau.data.Hedge;
import at.jku.risc.stout.hoau.data.TermNode;
import at.jku.risc.stout.hoau.data.atom.BoundVariable;
import at.jku.risc.stout.hoau.data.atom.Lambda;
import at.jku.risc.stout.hoau.data.atom.TermAtom;
import at.jku.risc.stout.hoau.data.atom.Variable;
import at.jku.risc.stout.hoau.util.DataStructureFactory;

/**
 * This class represents a rule based system for computing a bijection by
 * variable permutation between two terms. The terms will match under the
 * computed permutation. This algorithm is used as auxiliary task in
 * {@linkplain AntiUnifySystem}.<br>
 * The algorithm M is described in the paper:<br>
 * <a href="http://www.risc.jku.at/home/abaumgar/">Alexander Baumgartner</a>, <a
 * href="http://www.risc.jku.at/people/tkutsia/">Temur Kutsia</a>, <a
 * href="http://www.iiia.csic.es/~levy/">Jordi Levy</a> and <a
 * href="http://ima.udg.edu/~villaret/">Mateu Villaret</a>, A Variant of
 * Higher-Order Anti-Unification
 * 
 * @author Alexander Baumgartner
 */
public class PermEquivSystem {
	private Set<Variable> dom;
	private Set<Variable> ran;
	private Map<Variable, Variable> rho = DataStructureFactory.$.newMap();
	private Map<Variable, Variable> pi = DataStructureFactory.$.newMap();
	private EquationSystem<PermEquivProblem> problemSet;

	public PermEquivSystem() {
		this(new EquationSystem<PermEquivProblem>() {
			@Override
			public PermEquivProblem newEquation() {
				return new PermEquivProblem();
			}
		});
	}

	public PermEquivSystem(EquationSystem<PermEquivProblem> problemSet) {
		this.problemSet = problemSet;
	}

	/**
	 * Initializes the permutation computation.
	 */
	public void start(Set<? extends Variable> domain,
			Set<? extends Variable> range) {
		dom = DataStructureFactory.$.newSet(domain);
		ran = DataStructureFactory.$.newSet(range);
	}

	/**
	 * Adds a matching problem to the permuting matcher.
	 */
	public void addEquation(TermNode from, TermNode to) {
		problemSet.add(problemSet.newEquation(from.clone(), to.clone()));
	}

	/**
	 * Computes a permuting matcher for given matching problems (see
	 * {@linkplain #addEquation(TermNode, TermNode)}) and given domain/range
	 * (see {@linkplain #start(Set, Set)}). Returns null if a clash appears,
	 * otherwise a substitution will be returned.
	 */
	public Map<Variable, Variable> compute(DebugLevel debugLevel,
			PrintStream out) {
		while (!problemSet.isEmpty()) {
			debug(debugLevel, out);

			PermEquivProblem eq = problemSet.popLast();
			TermNode left = eq.getLeft();
			TermNode right = eq.getRight();
			TermAtom lAtom = pi.get(left.getAtom());
			if (lAtom == null)
				lAtom = left.getAtom();
			TermAtom rAtom = rho.get(right.getAtom());
			if (rAtom == null)
				rAtom = right.getAtom();
			Hedge lHedge = left.getHedge();
			Hedge rHedge = right.getHedge();

			// Abstraction
			if (lAtom instanceof Lambda || rAtom instanceof Lambda) {
				if (!(lAtom instanceof Lambda))
					left.etaExpand(((Lambda) rAtom).getBoundVar());
				else if (!(rAtom instanceof Lambda))
					right.etaExpand(((Lambda) lAtom).getBoundVar());
				BoundVariable lBound = ((Lambda) left.getAtom()).getBoundVar();
				BoundVariable rBound = ((Lambda) right.getAtom()).getBoundVar();
				eq.abstrLambda();
				rho.put(rBound, lBound);
				problemSet.add(eq);
				debug("ABS-M", debugLevel, out);
			}
			// Permutation
			else if (left.getAtom().getType() == right.getAtom().getType()
					&& lHedge.size() == rHedge.size()
					&& dom.contains(left.getAtom())
					&& ran.contains(right.getAtom())) {
				dom.remove(left.getAtom());
				ran.remove(right.getAtom());
				pi.put((Variable) left.getAtom(), (Variable) right.getAtom());
				for (int i = lHedge.size() - 1; i >= 0; i--)
					addEquation(lHedge.get(i), rHedge.get(i));
				debug("PER-M", debugLevel, out);
			}
			// Decomposition
			else if (lAtom.equals(rAtom) && lHedge.size() == rHedge.size()
					&& !dom.contains(left.getAtom())
					&& !ran.contains(right.getAtom())) {
				for (int i = lHedge.size() - 1; i >= 0; i--)
					addEquation(lHedge.get(i), rHedge.get(i));
				debug("DEC-M", debugLevel, out);
			}
			// Clash
			else
				return null;
		}
		debug(debugLevel, out);
		return pi;
	}

	private void debug(DebugLevel debugLevel, PrintStream out) {
		if (debugLevel == DebugLevel.PROGRESS
				|| debugLevel == DebugLevel.PROGRESSORIGIN) {
			printState(out);
			out.println();
		}
	}

	public EquationSystem<PermEquivProblem> getProblemSet() {
		return problemSet;
	}

	public void printState(PrintStream out) {
		out.println("Matching: " + problemSet);
		out.println(" Dom/Ran: " + dom + " / " + ran);
		out.println("  Rho/Pi: " + rho + " / " + pi);
	}

	private void debug(String rule, DebugLevel debugLevel, PrintStream out) {
		if (debugLevel == DebugLevel.PROGRESS
				|| debugLevel == DebugLevel.PROGRESSORIGIN) {
			out.println(rule + " ==> ");
		}
	}

	@Override
	public String toString() {
		return pi.toString();
	}

	public void reset() {
		rho.clear();
		pi.clear();
		problemSet.clear();
	}
}
